/**************************************************************************
*************** CALCULO DE LA SEMANTICA DE PUNTO FIJO: F(E). ************
***************************************************************************
* FICHERO fe.c: Calcula la semantica de punto fijo de una teoria, es decir
* el conjunto de ecuaciones ( no necesariamente pertenecientes a la base de
* Herbrand), que son teoremas de la teoria. El unico punto de entrada es la
* funcion:
*
*	fe(); 	Calcula la semantica de punto fijo valiendose del operador
*		TECA. El conjunto de ecuaciones de F(E) esta referenciado
*		por feini y fefin.
*
************************************************************************/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#include "const.h"
#include "estruc.h"
#include "glovar.h"

/* Prototipos de funciones externas, definidas en Procesos auxiliares.
No todas estas funciones son empleadas en "fe.c". */ 
#include "proaux.h"

/* Prototipos de funciones externas, definidas e Fichero de Unificacion
Sintactica. No todas estas funciones son empleadas en "fe.c". */
#include "unific.h"


/****************************  VISUALIZACION ****************************/

/* Impresion del conjunto de ecuaciones de una T-interpretacion */

PUBLIC verinter (struct stat *estado, int i)
{ struct arbol *ecuaux=NULL;

 fprintf(out,"\n\nTECA(%d)= ",i);
 for (;estado;estado=estado->sig)
   {
      ecuaux=estado->ecus->ecu; fprintf(out,"\n< ");
      if (ecuaux->oc[0]!='_') fprintf(out,"%s",ecuaux->oc);
      else fprintf(out,"%c%s",'X',&(ecuaux->oc[1]));
      if (ecuaux->hijo!=NULL)
	{ fprintf(out,"("); verterm(ecuaux->hijo); fprintf(out,")");  }
      fprintf (out,"  =  "); verterm(ecuaux->herm);
      fprintf(out,", ");verunif(estado->alfa);fprintf(out," >     Ec. asociada: ");
      ecuaux=copiecu(estado->ecus->ecu); aplicar(&ecuaux,estado->alfa);
      if (ecuaux->oc[0]!='_') fprintf(out,"%s",ecuaux->oc);
      else fprintf(out,"%c%s",'X',&(ecuaux->oc[1]));
      if (ecuaux->hijo!=NULL)
	{ fprintf(out,"("); verterm(ecuaux->hijo); fprintf(out,")");  }
      fprintf (out,"  =  "); verterm(ecuaux->herm); borraecu(ecuaux);
   }
}



/**************************  CALCULO DE FI(E)  *****************************/

/* Transforma un simbolo de funcion con su aridad, en una
ecuacion identidad, de la forma f(X,Y,..)=f(X,Y,..)  */

struct arbol *igualecu(struct funcs *simb)
{
 struct arbol *ecuaux1=NULL, *ecuaux2=NULL;
 char var[10];
 unsigned int i;

 if ((ecuaux1=(struct arbol *)malloc(sizeof(struct arbol)))==NULL)
			memerror();
 strcpy(ecuaux1->oc, simb->nombre); ecuaux1->herm=NULL;
 if (simb->aridad==0) ecuaux1->hijo=NULL;
 else
    {
      if ((ecuaux1->hijo=(struct arbol *)malloc(sizeof(struct arbol)))==NULL)
				memerror();
      ecuaux2=ecuaux1->hijo; fresh(var);
      strcpy(ecuaux2->oc,var);ecuaux2->hijo=NULL;
      for(i=1;i<(simb->aridad);i++)
	{
	  if ((ecuaux2->herm=(struct arbol *)malloc(sizeof(struct arbol)))==NULL)
					memerror();
	  ecuaux2=ecuaux2->herm;fresh(var);
	  strcpy(ecuaux2->oc,var);ecuaux2->hijo=NULL;
	}
      ecuaux2->herm=NULL;
    }
 ecuaux1->herm=copiterm(ecuaux1);
 return ecuaux1;
} 

/* Obtiene una lista de todas  las ecuaciones que pertenecen
a fi(E), es decir, el conjunto de todas las ecuaciones identidad
de la forma f(X,Y,..)=f(X,Y,...), para todos los posibles simbolos
de funcion f (de cualquier aridad) que aparezcan en la teoria E. */

struct lista *fie()
{                                          
 struct lista *listaux1=NULL, *listaux2=NULL;
 struct funcs *simbaux1=funcini;

 if (simbaux1!=NULL)
  {
    if ((listaux1=(struct lista *)malloc(sizeof(struct lista)))==NULL)
			memerror();
    listaux1->ecu=igualecu(simbaux1); listaux2=listaux1;
    listaux2->sig=NULL;
    simbaux1=simbaux1->sig; 
    for (;simbaux1!=NULL; )
      {
	 if ((listaux2->sig=(struct lista *)malloc(sizeof(struct lista)))==NULL)
				memerror();
	 listaux2=listaux2->sig; listaux2->ecu=igualecu(simbaux1);
	 listaux2->sig=NULL;  
	 simbaux1=simbaux1->sig; 
      }  
  }

 return listaux1;
}


/* Genera una lista de estados  iniciales de los que se parte para  
el calculo de la Semantica de Punto Fijo. Toma como argumento una 
lista de axiomas reflexivos y un doble puntero a estado inicializado 
a NULL que recoge el final de la lista de estados que se da como resultado.*/

PRIVATE struct stat *ltos_fie(struct lista *list, struct stat **fin)
{
 struct lista *lisfie=NULL;
 struct stat *estaux1=NULL, *estaux2=NULL, *estaux3=NULL;

lisfie=list;
while (lisfie)
   {
     if ((estaux2=(struct stat *)malloc(sizeof(struct stat)))==NULL) 
				memerror();
     if ((estaux2->ecus=(struct lista *)malloc(sizeof(struct lista)))==NULL)
				memerror();
     estaux2->ecus->ecu=copiecu(lisfie->ecu); /* solo es una ecuacion. */
     estaux2->ecus->sig=NULL;
     estaux2->raiz=NULL;
     estaux2->alfa=NULL; 
     estaux2->sig=NULL;
      /* enlazo el elemento en la lista de estados de narrowing */
     if (estaux1 == NULL) estaux1=estaux2;
     else estaux3->sig= estaux2;
     estaux3= estaux2;
     lisfie= lisfie->sig;
   }
   (*fin) = estaux3;
   return estaux1;
}


/* Transforma los estados de exito en el conjunto de ecs. F(E) */

PRIVATE stol_fe(struct stat *est)
{
   struct arbol *e;
   struct lista *l;
   struct stat  *s;

   for(s=est; s; s=s->sig) {
	if ((l= (struct lista *) malloc(sizeof(struct lista))) == NULL)
				memerror();
	e=copiecu(s->ecus->ecu);
	aplicar(&e, s->alfa);

	l->ecu=e; l->sig=NULL;
	if (feini == NULL) feini=l;
	else fefin->sig=l;
	fefin=l;
   }
}

/* Devuelve el numero de elementos de una lista de estados. */

int numelem(struct stat *s)
{    int n=0;
     for(;s;n++,s=s->sig);
     return n;
}

/************* UNIFICACION DE CONJUNTOS DE ECUACIONES *******************/

/* Aplica a todas las ecuaciones de una lista de estados de una
T-interpretacion sus correspondientes unificadores devolviendo una
lista de ecuaciones. No altera el parametro de entrada. */

PRIVATE struct lista *estalista(struct stat *s)
{
  struct lista *listaux1=NULL, *listaux2=NULL;
  struct stat *estaux=NULL;
  for(estaux=s; estaux; estaux=estaux->sig)
    { if (listaux1==NULL) 
	 { if ((listaux1=(struct lista *)malloc(sizeof(struct lista)))==NULL)
					     memerror();
	  listaux2=listaux1;}
      else
	 { if ((listaux2->sig=(struct lista *)malloc(sizeof(struct lista)))==NULL)
					     memerror();
	  listaux2=listaux2->sig;}
      listaux2->ecu=copiecu(estaux->ecus->ecu);
      aplicar(&(listaux2->ecu), estaux->alfa);
      listaux2->sig=NULL;
    }
  return listaux1;
}


/* Recoge el cuerpo de una regla y lo desglosa en una lista de 
ecuaciones. Altera el parametro de entrada*/

PRIVATE struct lista *bodylista(struct arbol *a)
{                             
  struct lista *listaux1=NULL, *listaux2=NULL;
  struct arbol *body=NULL, *ecuaux=NULL;
  for(body=a; body;)
    { if (listaux1==NULL) 
	 { if ((listaux1=(struct lista *)malloc(sizeof(struct lista)))==NULL)
					     memerror();
	  listaux2=listaux1;}
      else
	 { if ((listaux2->sig=(struct lista *)malloc(sizeof(struct lista)))==NULL)
					     memerror();
	  listaux2=listaux2->sig;}
      listaux2->ecu=body; body=body->herm->herm; 
      listaux2->ecu->herm->herm=NULL; listaux2->sig=NULL;
    }
  return listaux1;
}


/* Recibe una lista de ecuaciones e que intenta unificar con todos los
posibles subconjuntos de la lista de ecuaciones s. El resultado se
devuelve como una lista de unificadores apuntada por totalini al principio
y totalfin al final. Se pasa tambien el parametro mgu que se corresponde
con el unificador parcial que se esta calculando. */

PRIVATE unif(struct lista *e, struct lista *s, struct subs *mgu,
			   struct lisub **totalini, struct lisub **totalfin) 				
{ 
     int exito=0;
     struct arbol *ecuaux1=NULL, *ecuaux2=NULL;
     struct lista *listaux=NULL;
     struct lisub *lisini=NULL, *lisfin=NULL;
     struct subs *subaux1=NULL, *subaux2=NULL;

     ecuaux1=copiecu(e->ecu);
     aplicar(&ecuaux1,mgu);
     for(listaux=s; listaux; listaux=listaux->sig)
	{   ecuaux2=estandar(listaux->ecu);
	    aplicar(&ecuaux2,mgu);
	    subaux1=NULL;
	    if (unificador(ecuaux1, ecuaux2, &subaux1))
	      {	 if (e->sig)
		    { subaux2=componer(mgu,subaux1);
		      lisini=NULL;lisfin=NULL;
		      if (unif(e->sig, s , subaux2 , &lisini, &lisfin))
			  { 
			     if (*totalini) (*totalfin)->sig=lisini; 
			     else  (*totalini)=lisini;
			     (*totalfin)=lisfin;
			     exito=1;
			   }
		      borrasub(subaux2);
		     }
		 else
		     { if ((lisini=(struct lisub *)malloc(sizeof(struct lisub)))==NULL)
				      memerror();
		       lisini->alfa=componer(mgu,subaux1);
		       lisini->sig=NULL;
		       if (*totalini) (*totalfin)->sig=lisini; 
		       else  (*totalini)=lisini;
		       (*totalfin)=lisini;
		       exito=1;
		     }
	      }	   
	    borrasub(subaux1);
	    borraecu(ecuaux2);
	}
  borraecu(ecuaux1);
  return exito;
}


/* Parte de una lista de ecuaciones e (cuerpo de una regla
estandarizada) y un conjunto de ecuaciones que forman parte de una
T-interpretacion, y devuelve la lista de unificadores que existen entre
e y cualquier subconjunto de s. */

PRIVATE unifecus (struct lista *e, struct lista *s, struct lisub **total)
{
  struct lisub *lisini=NULL, *lisfin=NULL;
  struct lista *flate=NULL, *flatefin=NULL;

  *total=NULL;
  if ((e==NULL)) 
     {  if (((*total)=(struct lisub *)malloc(sizeof(struct lisub)))==NULL)
		       memerror();
	(*total)->alfa=NULL; (*total)->sig=NULL;
	return 1; 
     }
  else  if (s==NULL) return 0;
	else {  flate=flat(e, &flatefin);
		if (unif(flate,s,NULL,&lisini,&lisfin))
		   { *total=lisini; borralis(flate); return 1; }
		else {borralis(flate);return 0;}
	     }
}


/****************************  CALCULO DE TECA  **************************/

/* Aplicacion del operador TECA a una T-interpretacion (lista de estados) */

struct stat *TECA(struct stat *I)
{

 struct stat 	*estaux1=NULL, *estaux2=NULL, *estaux3=NULL;
 struct arbol 	*u=NULL, /* puntero a copia de una ocurrencia no var. de r. */
		*estapart=NULL,  /* Idem. ecuaci<=n estandarizada de la teoria, E. */
		*landa=NULL, /* Idem. parte izquierda de la ecuaci<=n estandarizada de E. */
		*ro=NULL, /* Idem. parte derecha de la ecuacion estandarizada. */
		*landaux=NULL,
		*uaux=NULL;
 struct lista   *listaux1=NULL, *listaux2=NULL, *listaux3=NULL, 
		*listaux4=NULL, *s=NULL, *e=NULL;
 struct subs 	*subaux1 =NULL, *subaux2=NULL;
 struct lisub 	*lisubaux1=NULL, *lisubaux2=NULL;

 s=estalista(I);
 for(estaux1=I; estaux1; estaux1=estaux1->sig)
   { /* para cada uno de los elementos de la T-interpretacion I. */
     st.BN_RU1_TECAt++;
     listaux1=ocubas(&(estaux1->ecus->ecu->herm),&listaux2);
     for(listaux2=listaux1;listaux2;listaux2=listaux2->sig)
       {/* para cada una de las ocurrencias no variables de r, del elemento <l=r, theta> de I. */
	u=copiterm(listaux2->ecu);
	aplicar(&u,estaux1->alfa);
	for (listaux3=listaini;listaux3;listaux3=listaux3->sig)
	  {/* para cada una de las ecuaciones de la teoria. */
	     estapart=estandar(listaux3->ecu);
	     landa=estapart;
	     ro=estapart->herm;
	     e=bodylista(estapart->herm->herm);
	     landa->herm=NULL;
	     ro->herm=NULL;
	     if (unifecus(e,s,&lisubaux1))
	       for (lisubaux2=lisubaux1;lisubaux2;lisubaux2=lisubaux2->sig)
		{  
		   landaux=copiterm(landa);
		   aplicar(&landaux,lisubaux2->alfa);
		   uaux=copiterm(u);
		   aplicar(&uaux,lisubaux2->alfa);
		   subaux1=NULL;
		   if (unificador(landaux,uaux,&subaux1))
		     {
		       st.BN_RU1_TECAe++;
		       if (estaux2==NULL)
			 {
			   if ((estaux2=(struct stat *)malloc(sizeof(struct stat)))==NULL)
					     memerror();
			   estaux3=estaux2;
			 }
		       else
			 {
			   if ((estaux3->sig=(struct stat *)malloc(sizeof(struct stat)))==NULL)
					     memerror();
			   estaux3=estaux3->sig;
			  }
			 /* Preparacion de la lista de ecuaciones del nuevo estado de la 
			  T-interpretacion. Las listas son de una sola ecuacion. */
		       if ((listaux4=(struct lista *)malloc(sizeof(struct lista)))==NULL)
					     memerror();
		       listaux4->ecu=modifica(estaux1->ecus->ecu, listaux2->ecu, ro);
		       listaux4->sig=NULL;
		       /* Completar el estado. */
		       estaux3->ecus=listaux4;
		       estaux3->raiz=NULL;
		       subaux2=componer(lisubaux2->alfa, subaux1);
		       estaux3->alfa=componer(estaux1->alfa,subaux2);
		       borrasub(subaux2);
		       estaux3->sig=NULL;
		     }
		   borraecu(landaux);
		   borraecu(uaux);
		   borrasub(subaux1);
		}
	     /* Con los siguientes tres borrados queda eliminada la
		ecuacion estandarizada por completo. */
	     borraecu(landa);
	     borraecu(ro);
	     borralis(e);
	     borralisub(lisubaux1);
	  }
	borraecu(u);
       }
     borraocu(listaux1);
   }
 /*borralis(s);*/
 return(estaux2);
}

/************ CALCULO DE LA SEMANTICA DE PUNTO FIJO: F(E) ***************
el calculo de F(E) consistirss en los siguientes pasos:

a) 	Borrar la lista, feini, que almacena los calculos de F(E) anteriormente realizados. Se desea rehacer los csslculos o bien se llama al procedimiento porque se introdujo una nueva teoroa; en cualquier caso hay que borrar los anteriores.

b) 	Generar la lista de ecuaciones, FI(E), de axiomas reflexivos, asociada a la teoroa.

c) 	Preparar la lista de estados, FIE, asociados a la teoria; un estado por cada ecuaci<=n de la lista FI(E). El campo oecusoe apuntarss a una lista formada por una sola ecuaci<=n de FI(E); el campo oraizoe no tiene utilizaci<=n, y se hace apuntar a NU

d) 	Aplicar reiteradamente el operador TECA a la T-interpretacion (conjunto de estados) anterior hasta alcanzar el punto fijo; El csslculo comienza aplicando TECA(FIE).

e) 	Transformar la lista de estados resultado en la lista que representa la semantica de punto fijo de la teoroa, F(E). Esta lista esta referenciada por los punteros feini y fefin. Se podra visualizar posteriormente con verlis().

**************************************************************************/

PUBLIC fe( )
{
	int i=0;
	struct lista *listaux1=NULL; 
	struct stat *FIE=NULL, *S=NULL, /* Almacena estados de exito obtenidos en cada  iteracion TECA*/
		    *Saux1=NULL, *Saux2=NULL, *Saux3=NULL;

	/* Se borran estructuras asociadas con el calculo de F(E) */
	start_st(PUNTOFIJO);
	if (feini) { borralis(feini); feini=NULL; fefin=NULL; }

	FIE = ltos_fie(fie(), &Saux3); /* Disgrega la lista de
		ecuaciones reflexivas y forma una de estados  FI(E).  */

	/* Calculo de los estados que forman la semantica operacional. */
	S=copistat(FIE);
	if (trace) verinter(S, i); /* visualiza TECA(0) si la traza esta activada. */
	for(;;) {    /* Aplica reiteradamente el operador TECA. */
	   Saux1=TECA(S);
	   Saux2=copistat(FIE);
	   if (Saux1)
	      {  for(Saux3=Saux2;Saux3->sig;Saux3=Saux3->sig);
		 Saux3->sig=Saux1; }
	   Saux1=Saux2;
	   if (numelem(S)==numelem(Saux1)) break; /* se ha alcanzado el punto fijo */
	   i++;  borraTinter(S);
	   if (trace) verinter(Saux1, i); /* visualiza TECA(i) si la traza esta activada. */
	   S=Saux1;
	   }
	stol_fe(Saux1); borraTinter(S); borraTinter(Saux1);
	if (trace)  fprintf (out,"\nPUNTO FIJO ALCANZADO.\n");
	stop_st();
}

