/************************************************************************
**************** CALCULO DE LA SEMANTICA OPERACIONAL: O(E).  ************
*************************************************************************
* FICHERO oe.c: Calcula la semantica operacional de una teoria, es decir
* el conjunto de ecuaciones FLAT (no necesariamente pertenecientes a la
* base de Herbrand), que son teoremas de la teoria. El unico punto de
* entrada es la funcion:
*
*	oe(); Calcula la semantica operacional. Devuelve el conjunto de
*	      ecuaciones de O(E).
*
************************************************************************/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#include "const.h"
#include "estruc.h"
#include "glovar.h"

/* Prototipos de funciones externas, definidas en Procesos auxiliares.
No todas estas funciones son empleadas en "oe.c". */ 
#include "proaux.h"

/* Prototipos de funciones externas, definidas e Fichero de Unificacion
Sintactica. No todas estas funciones son empleadas en "oe.c". */
#include "unific.h"

/* Prototipos de funciones externas, definidas e Fichero de Unificacion
Semantica. No todas estas funciones son empleadas en "oe.c". */
#include "narrow.h"


/****   GENERACION DE ECUACIONES FLAT Y ESTADOS DE NARROWING INICIALES  ****/


/* Transforma un simbolo de funcion con su aridad, en una
ecuacion plana */

PRIVATE struct arbol *flatecu(struct funcs *simb)
{
 struct arbol *ecuaux1=NULL, *ecuaux2=NULL;
 unsigned char var[10];
 unsigned int i;

 if ((ecuaux1=(struct arbol *)malloc(sizeof(struct arbol)))==NULL)
			memerror();
 strcpy(ecuaux1->oc, simb->nombre);
 if (simb->aridad==0) ecuaux1->hijo=NULL;
 else
    {
      if ((ecuaux1->hijo=(struct arbol *)malloc(sizeof(struct arbol)))==NULL)
				memerror();
      ecuaux2=ecuaux1->hijo; fresh((char *) var);
      strcpy(ecuaux2->oc, (char *) var);ecuaux2->hijo=NULL;
      for(i=1;i<(simb->aridad);i++)
	{
	  if ((ecuaux2->herm=(struct arbol *)malloc(sizeof(struct arbol)))==NULL)
					memerror();
	  ecuaux2=ecuaux2->herm;fresh((char *) var);
	  strcpy(ecuaux2->oc,(char *) var);ecuaux2->hijo=NULL;
	}
      ecuaux2->herm=NULL;
    }
 if ((ecuaux1->herm=(struct arbol *)malloc(sizeof(struct arbol)))==NULL)
				memerror();
 ecuaux1->herm->herm=NULL; ecuaux1->herm->hijo=NULL;
 fresh((char *) var); strcpy(ecuaux1->herm->oc,(char *) var);
 return ecuaux1;
} 


/* Obtiene una lista de todas las posibles ecuaciones flat que se
pueden formar con las funciones de la teoria */

PRIVATE struct lista *flatteo( )
{                                          
 struct lista *listaux1=NULL, *listaux2=NULL;
struct funcs *simbaux1= funcini;

 if (simbaux1!=NULL)
  {
    if ((listaux1=(struct lista *)malloc(sizeof(struct lista)))==NULL)
			memerror();
    listaux1->ecu=flatecu(simbaux1); listaux2=listaux1;
    listaux2->sig=NULL;
    simbaux1=simbaux1->sig; 
    for (;simbaux1!=NULL; )
      {
	 if ((listaux2->sig=(struct lista *)malloc(sizeof(struct lista)))==NULL)
				memerror();
	 listaux2=listaux2->sig;listaux2->ecu=flatecu(simbaux1);
	 listaux2->sig=NULL;  
	 simbaux1=simbaux1->sig;
      }  
  }

 return listaux1;
}

/* Genera una lista de estados de narrowing iniciales de los que se parte para 
el calculo de la Semantica Operacional. Toma como argumento una lista de 
ecuaciones FLAT, da como resultado la lista de estados de narrowing , ya 
mencionada. Como efecto lateral disgrega la lista de ecuaciones FLAT. */

PRIVATE struct stat *ltos_flat(struct lista *lisflat)
{
 struct lista *listaux2=NULL;
 struct stat *estaux1=NULL, *estaux2=NULL, *estaux3=NULL;

while (lisflat)
   {
     /* Individualiza el elemento de cabeza de la lista de ecuaciones flat. */
     listaux2= lisflat; lisflat= lisflat->sig;
     listaux2->sig= NULL;
    /* Creo un elemento de la lista de estados de narrowing */
     if ((estaux2=(struct stat *)malloc(sizeof(struct stat)))==NULL) 
				memerror();
     
     estaux2->ecus=listaux2; 
     estaux2->raiz=listaux2;
     estaux2->alfa=NULL;estaux2->sig=NULL;
      /* enlazo el elemento en la lista de estados de narrwing */
     if (estaux1 == NULL) estaux1=estaux2;
     else estaux3->sig= estaux2;
     estaux3= estaux2;
   }
   return estaux1;
}

/* Transforma los estados de exito en el conjunto de ecs. O(E) */
PRIVATE stol_oe(struct stat *est)
{
   struct arbol *e;
   struct lista *l;
   struct stat  *s;

   for(s=est; s; s=s->sig) {
	if ((l= (struct lista *) malloc(sizeof(struct lista))) == NULL)
				memerror();
	e=copiecu(s->raiz->ecu);
	aplicar(&e, s->alfa);

	l->ecu=e; l->sig=NULL;
	if (oeini == NULL) oeini=l;
	else oefin->sig=l;
	oefin=l;
   }
}

/*****************************************************************************
 Calculo de la semantica operacional, O(E). De acuerdo a su definicion formal:

O(E) = {(f(x1,..,xn) = xn+1).theta /  
		(f(x1,..,xn) = xn+1) ecuacion plana & 
		<<= f(x1,.., xn) = xn+1, ident> -BN->* <<=true, theta>};

el calculo de la misma consistir_ en los siguientes pasos:

A) 	Borrar la lista, oeini, que almacena los calculos anteriormente 
realizados. Se desea rehacer los calculos o bien se llama al procedimiento 
porque se introdujo una nueva teoria; en cualquier caso hay que borrar los 
anteriores.
B) 	Generar la lista de ecuaciones planas asociadas a la teoria.
C) 	Preparar la lista de estados de narrowing, E, asociados a la teor|a; un
estado por cada ecuacion plana. El campo ecus apuntara a una lista formada por 
una sola ecuacion, la ecuacion plana; lo mismo sucede con el campo raiz; el 
campo alfa apunta a la sustitucion identidad, es decir, sera NULL.
D) 	Aplicar basic narrowing al conjunto de estados anterior: S= narrower(E).
S apunta a una lista de estados de exito.
E) 	Transformar la lista de estados de exito en la lista que representa la 
semantica operacional de la teor|a, O(E). Esta lista esta referenciada por los 
punteros oeini y oefin.

En la programacion del procedimiento oe() se sigue esta linea por claridad. 
Podria haberse realizado mediante una composicion de funciones:

		PUBLIC oe() { stol_oe(narrower(ltos_flat(flatteo( )))); }

pero queda bastante criptico, y es mas dificil de depurar.
******************************************************************************/

PUBLIC oe( )
{
	struct lista *L; /* lista de ecuaciones flat */
	struct stat *E=NULL, /* lista estados narrowing asociados a la teoria*/
		   *S=NULL; /* lista estados  exito obtenidos por narrower() */

	start_st(OPERACIONAL);
	/* borro estructuras asociadas con el calculo de O(E) */
	if (oeini) { borralis(oeini); oeini=NULL; oefin=NULL; }

	L= flatteo();
	E= ltos_flat(L); /* Disgrega la lista L y forma una de estados
			    FLAT. No hay que borrar L. */
	S= narrower(E);
	borrastat(E);
	stol_oe(S);
	borrastat(S);
	stop_st();
}


