/**************************************************************************
************************ PROCESOS AUXILIARES ***************************
* FICHERO proaux.c: Este fichero contiene todos los procedimientos
* auxiliares empleados en mas de uno de los ficheros de la aplicacion.
* Casi todas las funciones en el definidas, son puntos de entrada a este
* fichero.
**************************************************************************/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <time.h>

#include "const.h"
#include "estruc.h"
#include "glovar.h"



/*********************  GESTION DE LA MEMORIA  *********************/

/* Deteccion de un error de memoria */

PUBLIC memerror()
 { printf("\nError. No queda memoria libre disponible. \n"); exit(1); }


/******************** Borrado de estructuras.  ******************/

/*elimina toda la memoria asociada a un arbol (ecuacion o termino) */

PUBLIC borraecu(struct arbol *ecu)
{ if (ecu) { 	borraecu(ecu->hijo); ecu->hijo=NULL;
		borraecu(ecu->herm); ecu->herm=NULL;
		free(ecu); ecu=NULL;}}


/*elimina toda la memoria asociada a una lista de ecuaciones. */

PUBLIC borralis(struct lista *lisini)
{
	struct lista *lsaux;
	while (lisini) { 
		lsaux=lisini; lisini=lisini->sig;
		borraecu(lsaux->ecu);
		lsaux->sig=NULL;
		free(lsaux);
	}
}

/*elimina toda la memoria asociada a una lista de simbolos.
Ahora solamente se utiliza en el procedimiento sintax.c, por lo
que podria ser un procedimiento pivado a ese modulo. Lo situamos
aqui por coherencia y si en el futuro pudiese requerirse en otros
modulos. */

PUBLIC borrafun (struct funcs *simbini)
{ if (simbini) {borrafun(simbini->sig);simbini->sig=NULL;
		free(simbini);simbini=NULL;};}


/*elimina toda la memoria asociada a una lista de ocurrencias.
En una lista de ocurrencias los terminos estan apuntados, por lo
que hay que tener cuidado a la hora de borrar sus elementos. Si
damos la orden de borrar el termino apuntado por el campo "ecu",
realmente estamos borrando un termino que forma parte de otra
estructura. */

PUBLIC borraocu (struct lista *ocuini)
{ if (ocuini) {	borraocu(ocuini->sig);ocuini->sig=NULL;
		free(ocuini);ocuini=NULL;};}


/*elimina toda la memoria asociada a una sustitucion */

PUBLIC borrasub (struct subs *alfa)
{ if (alfa) {	borrasub(alfa->sig);alfa->sig=NULL;
		borraecu(alfa->term);alfa->term=NULL;
		free(alfa);alfa=NULL;};}


/*elimina toda la memoria asociada a una lista de estados (de narrowing) */

PUBLIC borrastat (struct stat *est)
{ if (est) {
	borrastat(est->sig); est->sig=NULL;
	if (est->ecus!=est->raiz) { borralis(est->raiz); est->raiz=NULL;}
	borralis(est->ecus); est->ecus=NULL;
	borrasub(est->alfa); est->alfa=NULL;
	free(est);est=NULL;};
}
				     
/*elimina toda la memoria asociada a una T-interpretacion. */

PUBLIC borraTinter (struct stat *est)
{ if (est) {
	borraTinter(est->sig); est->sig=NULL;
	borralis(est->ecus); est->ecus=NULL;
	borrasub(est->alfa); est->alfa=NULL;
	free(est);est=NULL;};
}
				     

/*Elimina toda la memoria asociada a una lista de ocurrencias. En una lista de
ocurrencias los terminos estan apuntados, por lo que hay que tener cuidado a la
hora de borrar sus elementos. Si damos la orden de borrar el termino apuntado 
por el campo "term", realmente estamos borrando un termino que forma parte de 
otra estructura. */

PUBLIC borratocu (struct tocur *ocuini)
{ if (ocuini) {borratocu(ocuini->sig);free(ocuini);};}


/*Elimina toda la memoria asociada a una lista de substituciones o
unificadores. */

PUBLIC borralisub (struct lisub *listasub)
{ if (listasub) 
    { borralisub(listasub->sig); listasub->sig=NULL;
      borrasub(listasub->alfa); listasub->alfa=NULL;
      free(listasub); listasub=NULL;};
}


/***************** Copia de estructuras.  *********************/

/* Obtencion de una copia exacta de una ecuacion 
(teniendo en cuenta todos los hermanos ) */

PUBLIC struct arbol *copiecu (struct arbol *term)
 {
 struct arbol *ecuaux;
 if (term==NULL) return NULL;
 else
   {
    if ((ecuaux=(struct arbol *)malloc(sizeof(struct arbol)))==NULL)
				memerror();
    strcpy(ecuaux->oc, term->oc);
    ecuaux->hijo=copiecu(term->hijo); ecuaux->herm=copiecu(term->herm);
    return ecuaux;
   }
}


/* Obtencion de una copia exacta de un termino
 (por tanto, sin tener en cuenta sus hermanos)*/

PUBLIC struct arbol *copiterm (struct arbol *term)
{
 struct arbol *termaux;
 if (term==NULL) return NULL;
 else
   {
    if ((termaux=(struct arbol *)malloc(sizeof(struct arbol)))==NULL)
				memerror();
    strcpy(termaux->oc, term->oc);
    termaux->hijo=copiecu(term->hijo); termaux->herm=NULL;
    return termaux;
   }
}


/* Obtencion de una copia exacta de una lista de ecuaciones. */

PUBLIC struct lista *copilis(struct lista *lis)
{
   struct lista *listaux=NULL, *listaux1=NULL, *listaux2=NULL,
   		*listaux3=NULL;
   for (listaux1=lis; listaux1; listaux1=listaux1->sig) {
	if ((listaux2=(struct lista *) malloc(sizeof(struct lista))) == NULL) 
			memerror();
	listaux2->ecu=copiecu(listaux1->ecu);
	listaux2->sig=NULL;
	if (listaux==NULL) listaux=listaux2;
	else listaux3->sig=listaux2;
	listaux3=listaux2;
    }
    return listaux;
}


/* Obtencion de una copia exacta de una lista de substituciones. */

PUBLIC struct subs *copisub(struct subs *sub)
{
   struct subs  *listaux=NULL, *listaux1=NULL, *listaux2=NULL,
   		*listaux3=NULL;
   for (listaux1=sub; listaux1; listaux1=listaux1->sig) {
	if ((listaux2=(struct subs *) malloc(sizeof(struct subs))) == NULL) 
			memerror();
	strcpy(listaux2->var, listaux1->var);
	listaux2->term=copiterm(listaux1->term);
	listaux2->sig=NULL;
	if (listaux==NULL) listaux=listaux2;
	else listaux3->sig=listaux2;
	listaux3=listaux2;
    }
    return listaux;
}


/* Obtencion de una copia exacta de una lista de estados. */

PUBLIC struct stat *copistat(struct stat *est)
{
   struct stat  *listaux=NULL, *listaux1=NULL, *listaux2=NULL,
   		*listaux3=NULL;
   for (listaux1=est; listaux1; listaux1=listaux1->sig) {
	if ((listaux2=(struct stat *) malloc(sizeof(struct stat))) == NULL) 
			memerror();
	listaux2->ecus=copilis(listaux1->ecus);
	listaux2->raiz=copilis(listaux1->raiz);
	listaux2->alfa=copisub(listaux1->alfa);
	listaux2->sig=NULL;
	if (listaux==NULL) listaux=listaux2;
	else listaux3->sig=listaux2;
	listaux3=listaux2;
    }
    return listaux;
}

/***************************  VISUALIZACION **************************/

/* Impresion de un termino */

PUBLIC verterm(struct arbol *term)
{
 if (term)
  {
   if (term->oc[0]!='_') fprintf(out,"%s",term->oc);
   else fprintf(out,"%c%s",'X',&(term->oc[1]));
   if (term->hijo!=NULL) { fprintf(out,"("); verterm(term->hijo); fprintf(out,")");}
   if (term->herm!=NULL) { fprintf(out,","); verterm(term->herm); }
  }
}


/*Impresion de una ecuacion, sin mostrar mensajes */

PUBLIC verecu1(struct arbol *ecu)
{
  if (ecu)
    {
      if (ecu->oc[0]!='_') fprintf(out,"%s",ecu->oc);
      else fprintf(out,"%c%s",'X',&(ecu->oc[1]));
      if (ecu->hijo!=NULL){ fprintf(out,"("); verterm(ecu->hijo); fprintf(out,")");  }
      fprintf (out," = "); verterm(ecu->herm);
    }
}


/*Impresion de una ecuacion */

PUBLIC verecu(struct arbol *ecu)
{
  if (ecu)
    {
      fprintf(out,"\nEcuacion:      ");
      verecu1(ecu);
    }
}


/* Impresion de un conjunto de ecuaciones por pantalla. Toma como
argumentos un puntero al inicio de la lista y un mensaje que se mostrara
al principio. Por ejemplo si queremos visualizar las ecuaciones de la
semantica operacional escribiriamos : verlis("\n\nO(E): \n", oeini); */

PUBLIC verlis(char *mensa, struct lista *ini)
{  
   struct lista *laux;
   fprintf(out,mensa);
   for (laux=ini;laux;laux=laux->sig) verecu(laux->ecu);
}


/* Impresion del conjunto de reglas que forman una teoria por pantalla.*/ 

PUBLIC verteo()
{  
   struct lista *laux;
   struct arbol *ecuaux, *resto;
   fprintf(out,"\n\nTEORIA: \n");
   for (laux=listaini;laux;laux=laux->sig) 
	{ ecuaux=laux->ecu; 
	  resto=ecuaux->herm->herm;
	  ecuaux->herm->herm=NULL;
	  fprintf(out,"\nRegla:         ");
	  verecu1(ecuaux);
	  fprintf(out,"   <=   ");
	  if (resto)
	    { ecuaux->herm->herm=resto;
	      ecuaux=resto;
	      resto=ecuaux->herm->herm;
	      ecuaux->herm->herm=NULL;
	      verecu1(ecuaux);
	      while (resto)
		{ fprintf(out,",  ");
		  ecuaux->herm->herm=resto;
		  ecuaux=resto;
		  resto=ecuaux->herm->herm;
		  ecuaux->herm->herm=NULL;
		  verecu1(ecuaux);
		}
	    }
	}
}	    	  


/* Impresion de una lista de ocurrencias no variables por pantalla. Toma como
argumentos un puntero al inicio de la lista y un mensaje que se mostrara
al principio. Se empleo con fines de depuracion. */

PUBLIC verocu(char *mensa, struct lista *ini)
{  
   FILE *aux;
   struct lista *laux;
   aux=out; out=stdout;
   fprintf(out,mensa);
   for (laux=ini;laux;laux=laux->sig) {
		fprintf(out,"\n ocurencia: ");
		verterm(laux->ecu);};
   out=aux;
}


/* Visualiza todas las sustituciones de variables de un unificador */

PUBLIC verunif(struct subs *alfa)
{ struct subs *subaux;
  if (alfa==NULL) fprintf(out," E ");
  else
    for ( fprintf(out," { "), subaux=alfa;;)
      {
	 if (subaux->var[0]!='_') fprintf(out,"%s/",subaux->var);
	 else fprintf(out,"%c%s/",'X',&(subaux->var[1]));
	 verterm(subaux->term);
	 if (subaux->sig) { fprintf(out,", "); subaux=subaux->sig; }
	 else { fprintf(out," }"); break;}
      }
}


/* Visualiza todos los pares <listaecuaciones,sustitucion>
de una lista de estados */

PUBLIC verestado(struct stat *estado)
{
  for (;estado;estado=estado->sig)
   { printf("\nESTADO: ");
     verlis("\nEcuaciones:   ", estado->ecus);
     verlis("\nRaiz:         ", estado->raiz);
     printf("\nsubstitucion: "); verunif(estado->alfa);
   }
  printf("\nPulse RETURN para continuar ... "); getchar();
}


/**************************   OCURRENCIAS   *****************************/

/* Devuelve una ecuacion identica a la de partida pero sustituyendo
el termino viejo por el nuevo */

PUBLIC struct arbol *modifica (struct arbol *ecu, 
			struct arbol *viejo, struct arbol *nuevo)
{ struct arbol *ecuaux=NULL;

 if (ecu)
   { if(ecu==viejo)
	{
	   ecuaux=copiterm(nuevo);
	   ecuaux->herm=copiecu(ecu->herm);
	}
     else
	{
	  if ((ecuaux=(struct arbol*)malloc(sizeof(struct arbol)))==NULL)
			     memerror();
	   strcpy(ecuaux->oc,ecu->oc);
	   ecuaux->hijo=modifica(ecu->hijo,viejo,nuevo);
	   ecuaux->herm=modifica(ecu->herm,viejo,nuevo);
	}
   }
 return ecuaux;
}


/* Obtencion de una lista con todas las ocurrencias no variables
de una ecuacion */

PUBLIC struct lista *ocubas(struct arbol **ecu, struct lista **final)
{
 struct lista *listaux1=NULL, *listaux2=NULL,
	      *listaux3=NULL, *listaux4=NULL; 
 if ((*ecu))
  {
    if (((*ecu)->oc[0]!='_') && (!isupper((*ecu)->oc[0]))) /* Es una funcion */
       {  if ((listaux1=(struct lista *)malloc(sizeof(struct lista)))==NULL)
				memerror();
	  listaux1->ecu=(*ecu); listaux2=listaux1;
	  listaux2->sig=ocubas(&((*ecu)->hijo), &listaux3);
	  if (listaux2->sig) listaux2=listaux3;
       }
    listaux3=ocubas(&((*ecu)->herm), &listaux4);
    if (listaux3)
       if (listaux1)  { listaux2->sig=listaux3; listaux2=listaux4; }
       else {listaux1=listaux3;listaux2=listaux4;}
  }
 (*final)=listaux2;
 return listaux1;
}


/*************************   ESTANDARIZACION   **************************/

/*Devuelve un nombre de  variable "fresca" (no utilizada hasta ahora) */

PUBLIC fresh(char *var)
{ char varaux[9];
  int n;
  strcpy(var,"_");
  if ((n= (int) sprintf(varaux,"%-i",semilla)) > 8) {
	fprintf(out, "ERROR: valor semilla demasiado grande: %i%s%i\n",
		semilla, varaux, n);
	exit(1);
  };
  strcat(var,varaux); semilla++;
}


/* Cambia las apariciones de una variable "antigua" por otra "nueva"
dentro de una ecuacion */

PRIVATE cambiar(struct arbol *ecu, char *antigua, char *nueva)
{
  if (ecu)
   {
    if (!strcmp(ecu->oc,antigua)) strcpy(ecu->oc,nueva);
    cambiar(ecu->hijo,antigua,nueva);
    cambiar(ecu->herm,antigua,nueva);
   }
}


/* Indica si el nombre de una variable es mayor (ha
aparecido mas tarde) que el de otra */

PRIVATE mayor(char *a, char *b)
{ int l1,l2;
  if (b[0]!='_') return 1;
  else { l1=strlen(a); l2=strlen(b);
	 if (l1>l2) return 1;
	 else if (l1<l2) return 0;
	      else return (strcmp(a,b)>0);
	}
}


/* Refresca las variables (pone variables nuevas) en una ecuacion */

PRIVATE refrescar(struct arbol *ecu, struct arbol *term, char *primera)
{
 char varaux1[10], varaux2[10]; 
 if (term)
  {
    if ((isupper(term->oc[0])) || ((term->oc[0]=='_') && 
	(mayor(primera,term->oc)))) /* variable aun no refrescada */
	  {
	    strcpy(varaux1, term->oc); fresh(varaux2);
	    cambiar(ecu, varaux1, varaux2);
	  }
    refrescar(ecu,term->hijo,primera); 
    refrescar(ecu,term->herm,primera);
  }
}


/* Devuelve una ecuacion identica a la original, pero con variables
 frescas (standardised apart). */

PUBLIC struct arbol *estandar(struct arbol *ecu)
{
  struct arbol *ecuaux;
  char nueva[10];

  ecuaux=copiecu(ecu); fresh(nueva); semilla--;
  refrescar(ecuaux,ecuaux,nueva);
  return ecuaux;
}




/**************** CALCULO DE ECUACIONES PLANAS ******************************
*
*  Una ecuacion esta aplanada cuando tiene la forma f(x1, x2, ..., xN) = xN+1. 
* El objetivo de este apartado es conseguir que una expresion "s", que en el 
* caso mas general estara contituido por una lista de ecuaciones del objetivo 
* se tranforme en una lista de ecuaciones planas. El algoritmo de la funcion, 
* flat(s), que permite el aplanamiento de una expresion queda definido 
* inductivamente como:
*
* flat(s) =
*
*	1) flat(e1),..., flat(en)	Si s<= e1, ..., en;
*
*	2) flat(f(t1,...,tn)=z),
*		flat(g(tt1,...,ttn)=z)	Si s<= f(t1,...,tn)=g(tt1,...,ttn);
*					donde z es una variable fresca.
*
*	3) flat(f(t1,...,tn)=z)		Si s<= z=f(t1,...,tn);
*
*	4) flat(ti1=z1),...,flat(tik=zk),
*		flat(f(x1,...,xn)=z	Si s<= f(t1,...,tn)=z && ti1,...,tik 
*					son los argumentos no vatiables de 
*					f(t1,...,tn);
*					donde z1,...,zk son variables frescas.
*					(xi=zk si ti=tik) && (xi=ti si ti es 
*					variable)
*
*	5) f(x1,..., xn)=z, xn+1=z	Si s<= f(x1,..., xn)=xn+1 && Existe xi,
*					xi=xn+1;
*					donde z es una variable fresca.
*
*	6) xn=z, xn+1=z		Si s<= xn=xn+1 && xi es identica a xn+1;
*					donde z es una variable fresca.
*
*	7) s				En cualquier otro caso.
*
* Resolveremos el algoritmo mediante dos funciones: una flat(), que se encargara
* de tomar una lista de ecuaciones, e implementa el el caso (1); otra flatec(),
* que toma una sola ecuacion, e implementa los casos restantes.
*
**************************************************************************/

/* Comprueba si un subtermino es una variable. Toma como argumento un termino.
Devuelve un 1 si el termino es una variable, 0 en caso contrario. */

PUBLIC int esvar(struct arbol *term)
{
	if (term == NULL) return 0;
	return ( (isupper(term->oc[0])) || (term->oc[0] == '_') );
}


/* Comprueba si todos los argumentos, ti, de un termino f(t1, t2, ..., tN), o 
de la parte "izq" de una ecuacion, son o no variables. El argumento puede ser 
un termino funcion o una ecuacion. Retorna 2 si todos los ti son variables y 
alguno de ellos es igual al termino variable "der"  de la ecuaci n que se pasa 
como argumento. Retorna 1 si todos los ti son variables y ninguno de ellos es 
igual al termino "der" de la ecuacion. 
Retorna 0 en cualquier otro caso. Una ecuacion esta aplanada cuando tiene la 
forma f(x1, x2, ..., xN) = xN+1. Asi pues, esta funcion puede utilizarse para 
comprobar si una parte de la ecuacion ya esta en forma plana (retorna 2) o la 
ecuacion ya es plana (retorna 1).
La funcion se limita a recorrer los argumentos (Hermanos del hijo del termino 
apuntado) comprobando que son variables */

PRIVATE int esflat(struct arbol *term)
{
	int retorna=1;
	struct arbol *t;

	if (term == NULL) return 0;
	/* Esta linea se comenta por eficiencia:
	if (!esvar(term->herm)) return 0;
	Si no lo hacemos asi esvar(), se ejecuraria dos veces con el mismo 
	termino. Ver documentaci<=n */
	for (t=term->hijo; t; t=t->herm) {
		if (!esvar(t)) return 0;
		else if (strcmp(t->oc, term->herm->oc) == 0) retorna=2;
	}
	return retorna;
}


/* Toma un termino funcion no plano y genera una lista con las ocurrencias de 
argumentos no variables. Para faciliar posteriores operaciones de sustitucion 
de esos terminos por variables frescas se toma nota del termino que le precede 
en el campo "ant" de la estructura "tocur". */

PRIVATE struct tocur *term_novar(struct arbol *term)
{
	struct arbol *anterior, *t;
	struct tocur *l=NULL, *lfin=NULL, *l2;

	for (anterior=term, t=term->hijo; t; anterior=t, t=t->herm) {
	  if (!esvar(t)) {
		if ((l2=(struct tocur *) malloc(sizeof(struct tocur))) == NULL)
					memerror();
		l2->term=t; l2->ant=anterior; l2->sig=NULL;
		if (l==NULL) l=l2;
		else lfin->sig=l2;
		lfin=l2;
	  }
	}
	return(l);
}

/* Toma dos argumentos, el primero el puntero a una ecuacion a aplanar, el 
segundo un doble puntero al final de la lista de ecuaciones flat que obtenemos 
como resultado. Devuelve un puntero al inicio de la lista de ecuaciones flat. */

PRIVATE struct lista *flatec(struct arbol *ecu, struct lista **final)
{
	struct arbol *izq=NULL, *der=NULL, *z=NULL, *z2=NULL;
	struct lista *lfaux1=NULL, *lfaux2; /* Principio y fin lista ecus. 
					       flat intermedia */
	struct lista *lfaux3; /* auxiliar para concatenar listas ecus. 
				flat intermedias */
	struct tocur *l=NULL, *larg=NULL;
	izq=ecu; der=ecu->herm;
	
	if (!esvar(izq) && !esvar(der))
	{ /* CASO 2) Si f(t1,...,tn)=g(tt1,...,ttn) */

		/* crear la variable fresca z */
		if ((z= (struct arbol *) malloc(sizeof(struct arbol))) == NULL)
						memerror();
		fresh(z->oc); z->hijo=NULL; z->herm=NULL;

		/* crear las ecuaciones f(t1,...,tn)=z && g(tt1,...,ttn)=z */
		izq->herm = z;
		der->herm = copiterm(z);

		/*  crear lista de ecuaciones [flat(f(t1,...,tn)=z) && 
		flat(g(tt1,...,ttn)=z)] */
		lfaux1= flatec(izq, &lfaux2);
		lfaux3=lfaux2;
		lfaux3->sig= flatec(der, &lfaux2);
		(*final)=lfaux2;
		return lfaux1;
	}
	else if (esvar(izq) && !esvar(der))
	{/* CASO 3)	Si s<= z=f(t1,...,tn) */
		/* Crear la ecuacion f(t1,...,tn)=z cambiando posicion 
		izq por der */
		der->herm= izq; izq->herm=NULL;

		/* retornar la lista  flat(f(t1,...,tn)=z) */
		lfaux1=flatec(der, &lfaux2);
		(*final)=lfaux2;
		return (lfaux1);
	 }
	 else if (!esvar(izq) && esvar(der) && !esflat(izq))
	 { /*	CASO 4) Si s<= f(t1,...,tn)=z && ti1,...,tik son
	    los argumentos no vatiables de f(t1,...,tn) */
		/* lista de argumentos no variables */
		larg= term_novar(izq);
		for (l=larg; l; l=l->sig) {
		  /* crear la variable fresca zk */
		  if ((z= (struct arbol *) malloc(sizeof(struct arbol)))== NULL)
					memerror();
		  fresh(z->oc); z->hijo=NULL; z->herm=NULL;
		  z2=copiterm(z);
		  /* Aprovechamos para sustituir el termino tik de f(t1,...,tn)
		     por la variable zk */
		  if (ecu == l->ant)  /*  el arg. t1 es no variable */
			ecu->hijo=z; 
		  else /* el arg. es un ti intermedio o el ultimo arg. */
			l->ant->herm=z;
		  z->herm=l->term->herm; 
		  /* crear cada una de las ecuaciones tik=zk */
		  l->term->herm=z2;
		  /* anyadirlo a la lista de ecuaciones flat */
		  if (lfaux1==NULL) {
			lfaux1=flatec(l->term, &lfaux2); lfaux3=lfaux2;
		  }
		  else {
			lfaux3->sig=flatec(l->term, &lfaux2);lfaux3=lfaux2;
		  }
		}
		/* despues de este trabajo "ecu" referencia f(x1,...xn)=z
		hay que anyadirlo a la lista de ecuaciones flat */
		lfaux3->sig=flatec(ecu, &lfaux2);
		(*final)=lfaux2;
		borratocu(larg);
		return(lfaux1);
	  }
	  else if (!esvar(izq) && esvar(der) && (esflat(ecu)==2))
	  { /*	CASO 5) f(x1,..., xn)=xn+1 && Existe xi, xi=xn+1*/
		/* crear la variable fresca z */
		if ((z= (struct arbol *) malloc(sizeof(struct arbol))) == NULL)
				memerror();
		fresh(z->oc); z->hijo=NULL; z->herm=NULL;
		/* crear lista ecuaciones f(x1,..., xn)=z && xn+1=z . Estas
		ecuaciones ya son flat. */
		izq->herm=z; der->herm=copiterm(z);
		if ((lfaux1= (struct lista*) malloc(sizeof(struct lista))) == NULL) 
				memerror();
		lfaux1->ecu=izq;
		if ((lfaux2=(struct lista*) malloc(sizeof(struct lista))) == NULL) 
				memerror();
		lfaux2->ecu=der;  lfaux2->sig=NULL;
		lfaux1->sig=lfaux2;
		(*final)=lfaux2;
		return lfaux1;
	   }
	   else if (esvar(izq) && esvar(der) && !strcmp(izq->oc, der->oc))
	   { /*	CASO 6) xn=xn+1 && xi es identica a xn+1 */	
		/* crear la variable fresca z */
		if ((z= (struct arbol *) malloc(sizeof(struct arbol))) == NULL)	
				memerror();
		fresh(z->oc); z->hijo=NULL; z->herm=NULL;
		/* crear  ecuaciones xn=z && xn+1=z . Estas ecuaciones 
		ya son flat. */
		izq->herm=z; der->herm=copiterm(z);
		/* crear lista de ecuaciones flat */
		if ((lfaux1= (struct lista*) malloc(sizeof(struct lista))) == NULL)
				 memerror();
		lfaux1->ecu=izq;
		if ((lfaux2=(struct lista*) malloc(sizeof(struct lista))) == NULL)
				 memerror();
		lfaux2->ecu=der;  lfaux2->sig=NULL;
		lfaux1->sig=lfaux2;
		(*final)=lfaux2;
		return lfaux1;
	}
	else { /* En cualquier otro caso, la ecu es flat */
		if ((lfaux1= (struct lista*) malloc(sizeof(struct lista))) == NULL) 
				memerror();
		lfaux1->ecu=ecu; lfaux1->sig=NULL;
		(*final)=lfaux1;
		return lfaux1;
	 }
}

/* Toma una lista de ecuaciones y devuelve la lista de ecuaciones planas. Hace
uso de flatec(). Implementa el caso (1) del algoritmo de aplanamiento 
anteriormente descrito. */

PUBLIC struct lista *flat(struct lista *obj, struct lista **final)
{
	struct lista *l=NULL, *lfinaux=NULL, *lflatini=NULL, *lflatfin=NULL;
	l=copilis(obj);
	for (; l; l=l->sig) {
	  if (lflatini == NULL) lflatini=flatec(l->ecu, &lfinaux);
	  else lflatfin->sig=flatec(l->ecu, &lfinaux);
	  lflatfin=lfinaux; lfinaux=NULL;
	}
	(*final)=lflatfin; 
	return (lflatini);
}




/*************** ESTADISTICAS Y CALCULOS DE EFICIENCIA ****************/

PUBLIC start_st(char metodo)
{ 	/* Inicializar la variable st. */
	st.metodo=metodo;
	st.instan=0;
	st.compos=0;
	st.unifict=0;
	st.unifice=0;
	st.BN_RU1_TECAt=0;
	st.BN_RU1_TECAe=0;
	st.BN_RU2t=0;
	st.BN_RU2e=0;
	st.time=clock();
}


PUBLIC stop_st()
{
	st.time=clock()-st.time;
}

PUBLIC ver_st()
{
printf("\n\n\n\n\n\n\n\n\n");
printf("Ultimo calculo recogido en la estadistica:                          \n");
switch (st.metodo)
{
case OPERACIONAL:
	  printf("SEMANTICA OPERACIONAL DE UNA TEORIA                       \n\n");
	  printf ("Numero total de operaciones BN1:  \t\t\t%d\n", st.BN_RU1_TECAt);
	  printf ("Numero de operaciones BN1 con exito:  \t\t\t%d\n", st.BN_RU1_TECAe);
	  printf ("Numero total de operaciones BN2:  \t\t\t%d\n", st.BN_RU2t);
	  printf ("Numero de operaciones BN2 con exito:  \t\t\t%d\n", st.BN_RU2e);
	  break;
case PUNTOFIJO:
	  printf("SEMANTICA DE PUNTO FIJO DE UNA TEORIA                     \n\n");
	  printf ("Numero total de operaciones TECA:  \t\t\t%d\n", st.BN_RU1_TECAt);
	  printf ("Numero de operaciones TECA con exito:  \t\t\t%d\n", st.BN_RU1_TECAe);
	  break;
case CLASICO:
	  printf("RESOLUCION DE UN OBJETIVO POR EL METODO CLASICO           \n\n");
	  printf ("Numero total de operaciones BN1:  \t\t\t%d\n", st.BN_RU1_TECAt);
	  printf ("Numero de operaciones BN1 con exito:  \t\t\t%d\n", st.BN_RU1_TECAe);
	  printf ("Numero total de operaciones BN2:  \t\t\t%d\n", st.BN_RU2t);
	  printf ("Numero de operaciones BN2 con exito:  \t\t\t%d\n", st.BN_RU2e);
	  break;
case SEMANTICO:
	  printf("RESOLUCION DE UN OBJETIVO POR EL METODO SEMANTICO         \n\n");
	  printf ("Numero total de operaciones RU1:  \t\t\t%d\n", st.BN_RU1_TECAt);
	  printf ("Numero de operaciones RU1 con exito:  \t\t\t%d\n", st.BN_RU1_TECAe);
	  printf ("Numero total de operaciones RU2:  \t\t\t%d\n", st.BN_RU2t);
	  printf ("Numero de operaciones RU2 con exito:  \t\t\t%d\n", st.BN_RU2e);
	  break;
default : printf("NINGUN CALCULO");return;
}
printf ("Numero total de operaciones de instanciacion:  \t\t%d\n", st.instan);
printf ("Numero total de operaciones de composicion  :  \t\t%d\n", st.compos);
printf ("Numero total de operaciones de unificacion  :  \t\t%d\n", st.unifict);
printf ("Numero de operaciones de unificacion con exito :  \t%d\n", st.unifice);
printf ("Tiempo total empleado en el calculo:  \t\t\t%d\n\n", st.time);
printf("\n\n\n\n\n\n");
}




