/**************************************************************************
**************************  ANALISIS SINTACTICO **************************
***************************************************************************
* FICHERO sintax.c: En este fichero se realizan todas las labores sintacticas.
* Analiza los terminos, ecuaciones, reglas, teorias, y objetivos. Genera tres
* estructuras estables: La lista de reglas que forman la teoria; la lista de
* simbolos de la teoria (signatura); la lista de ecuaciones que forman un
* objetivo. Los unicos puntos de entrada para este fichero son:
*
*		parser_teo();  Permite introducir las reglas que
*			      forman la teoria.
*		parser_goal(); Desempena el mismo papel respecto a los
*			      objetivos.
*
**************************************************************************/
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#include "const.h"
#include "estruc.h"
#include "glovar.h"

#define	LCAD	81 /* Longitud buffer intermedio cad[]. */

/* Buffer interno para almacenar la entrada que se pretende analizar
e indices para recorrerlo. */
PRIVATE unsigned char cad[LCAD];
PRIVATE int i,j;

/* Prototipos de funciones externas, definidas en Procesos auxiliares. No
todas estas funciones son empleadas en "sintax.c". */
#include "proaux.h"

PRIVATE struct arbol *subterm();
PRIVATE struct funcs *ecufuncs(struct arbol *ecu, struct funcs **final);

/* Deteccion de un error sintactico */

PRIVATE sintaxerror()
 { printf("\nError sintactico en la posicion %d (%c) \n",i,cad[i]);
   exit(1); }


/* Saltar blancos */

PRIVATE sb()
{	for(;cad[i]==' '; i++);}


/* Analisis sintactico de un termino */

PRIVATE struct arbol *term ()
{
 struct arbol *termaux; int j;
 sb();
 if(cad[i]=='(')
   {
     i++;
     termaux=term();
     sb();
     if (cad[i]!=')') sintaxerror();
     i++;
   }
  else if (isalnum(cad[i]))
	{
	   if ((termaux=(struct arbol *)malloc(sizeof(struct arbol)))==NULL)
				memerror();
	   for(j=0,termaux->oc[j]=cad[i],j++,i++; (isalnum(cad[i]))&&(j<9);
		   termaux->oc[j]=cad[i],j++,i++);
	   for(termaux->oc[j]=0;isalnum(cad[i]) && (i<90); i++);
	   termaux->herm=NULL; termaux->hijo=NULL;
	   sb();
	   if (cad[i]=='(')
		{
		  i++;
		  termaux->hijo=subterm();
		  sb();
		  if (cad[i]!=')') sintaxerror();
		  i++;
		}
	}
      else sintaxerror();
  return termaux;
}


/* Analisis sintactico de una lista de subterminos */

PRIVATE struct arbol *subterm()
{
 struct arbol *ecuaux;
 sb();
 ecuaux=term();
 sb();
 if(cad[i]==',')
   {
     i++;
     ecuaux->herm=subterm();
   };
 return ecuaux;
}


/* Analisis sintactico de una ecuacion: TERM = TERM */

PRIVATE struct arbol *ecuacion()
{
 struct arbol *termaux;
 sb();
 termaux=term();
 sb();
 if (cad[i]!='=') sintaxerror();
 i++;
 termaux->herm=term();
 return termaux;
}


/* Analisis sintactico de una secuencia de ecuaciones separadas por comas. */

PRIVATE struct arbol *body()
{
 struct arbol *ecuaux1, *ecuaux2;
 sb();
 ecuaux1=ecuacion();
 sb();
 ecuaux2=ecuaux1;
 while (cad[i]==',')
   {
     i++;
     ecuaux2->herm->herm=ecuacion();
     ecuaux2=ecuaux2->herm->herm;
   };
 return ecuaux1;
}

/* Analisis sintactico de una regla. Una regla tiene el siguiente
formato: CABEZA <= CUERPO, donde CABEZA es una ecuacion, y CUERPO
es una secuencia posiblemente vacia de ecuaciones separadas por 
comas.  La funcion analiza la cadena cad[]. Los terminadores de una 
regla son los caracteres: '\0' , '\n'. */

PRIVATE struct arbol *regla()
{
 struct arbol *reglaux;
 sb();
 if ((cad[i]==0)||(cad[i]=='\n')) return NULL;
 reglaux=ecuacion();
 sb();
 if (cad[i]!='<') sintaxerror();
 i++;
 sb();
 if (cad[i]!='=') sintaxerror();
 i++;
 sb();
 if ((cad[i]!=0)&&(cad[i]!='\n')) reglaux->herm->herm=body();
 sb();
 if ((cad[i]!=0)&&(cad[i]!='\n')) sintaxerror();
 return reglaux;
}


/* Analiza la ultima regla leida y la inserta en la
lista global que enlaza todas las reglas de la teoria. */

PRIVATE tratarteo(struct lista **ini, struct lista **fin)
{
  struct arbol *reglaux=NULL;

  if ((reglaux=regla())!=NULL)
     {
       if ((*ini)==NULL)
	{
	 if (((*ini)=(struct lista *)malloc(sizeof(struct lista)))==NULL)
				memerror();
	 (*fin)=(*ini);
	}
       else
	{
	 if (((*fin)->sig=(struct lista *)malloc(sizeof(struct lista)))==NULL)
				memerror();
	 (*fin)=(*fin)->sig;
	};
       (*fin)->ecu=reglaux;(*fin)->sig=NULL;
      }
}


/* Analiza el ultimo objetivo leido y lo inserta en la
lista global que enlaza todas las ecuaciones del objetivo. 
Los goal tiene que comenzar con "<=", despues de ese simbolo
sigue una sucesion de ecuaciones separadas por ",". */
  
PRIVATE tratargoal(struct lista **ini, struct lista **fin)
{
   sb();
   if (cad[i]!='<') sintaxerror();
   i++;
   sb();
   if (cad[i]!='=') sintaxerror();
   i++;
   if (((*ini)=(struct lista *)malloc(sizeof(struct lista)))==NULL)
			memerror();
   (*fin)=(*ini);
   (*fin)->ecu=ecuacion();
   sb();
   while (cad[i]==',')
      { i++;
	if (((*fin)->sig=(struct lista *)malloc(sizeof(struct lista)))==NULL)
			memerror();
	(*fin)=(*fin)->sig;
	(*fin)->ecu=ecuacion();
	sb();
      };
   (*fin)->sig=NULL;
   if ((cad[i]!=0)&&(cad[i]!='\n')) sintaxerror();
}

/****************** GENERACION DE LA LISTA DE SIMBOLOS *****************
* Las funciones que se definen a continuacion tienen por objeto la
* generacion de la signatura asociada a la teoria. El principal papel que
* damos aqui a esta estructura es sintactico, pues nos servira para realizar
* un control de errores sintacticos adicionales.
**************************************************************************/


/* Devuelve 1 si un determinado simbolo de funcion se encuentra 
en una lista de simbolos (0 en otro caso). Dos simbolos se consideran
iguales si tienen el mismo nombre y aridad. Asi pues podran existir en la
lista de simbolos, que generemos, nodos con simbolos de igual nombre y
diferente aridad. Esto ultimo debera detectarse como error sintactico y
comunicarse al usuario. */

PRIVATE esta(struct funcs *listsimb, struct funcs *simb)
{
 for (;listsimb!=NULL;listsimb=listsimb->sig)
     if (!strcmp(listsimb->nombre,simb->nombre) && 
	 (listsimb->aridad==simb->aridad)) break;
 if (listsimb!=NULL) return 1;
 else return 0;
}

/* Devuelve 1 si un determinado simbolo de funcion se encuentra 
en una lista de simbolos (0 en otro caso). Dos simbolos se consideran
iguales si tienen el mismo nombre. Asi pues, detectara las
apariciones de simbolos con igual nombre y diferente aridad. Esto 
ultimo debera detectarse como error sintactico y comunicarse al 
usuario. */

PRIVATE esta1(struct funcs *listsimb, struct funcs *simb)
{
 for (;listsimb!=NULL;listsimb=listsimb->sig)
     if (!strcmp(listsimb->nombre,simb->nombre)) break;
 if (listsimb!=NULL) return 1;
 else return 0;
}


/* Partiendo de un termino, devuelve una lista de todas las 
funciones con sus aridades, que aparecen en el mismo. Puede haber
nombres de simbolos repetidos con diferentes aridades. */

PRIVATE
struct funcs *termfuncs(struct arbol *term, struct funcs **final)
{
 struct arbol *termaux;
 struct funcs *simbaux1=NULL, *simbaux2=NULL, 
	      *simbaux3=NULL, *simbaux4=NULL;
 unsigned int i=0;

 if (term!=NULL)
  {
    if (!isupper(term->oc[0])) /*simbolo de funcion */
      {
	if ((simbaux1=(struct funcs *)malloc(sizeof(struct funcs)))==NULL)
				memerror();
	simbaux2=simbaux1;
	simbaux2->sig=NULL;
	strcpy(simbaux2->nombre,term->oc);
	for(i=0,termaux=term->hijo; termaux!=NULL; termaux=termaux->herm,i++);
	simbaux2->aridad=i;
	simbaux2->sig=termfuncs(term->hijo,&simbaux3);
	if (simbaux2->sig!=NULL) simbaux2=simbaux3;
      }

      simbaux3=termfuncs(term->herm,&simbaux4);

      if (simbaux3!=NULL)
			{
				if (simbaux1==NULL) simbaux1=simbaux3;
				else simbaux2->sig=simbaux3;
				simbaux2=simbaux4;
			}
  }
 (*final)=simbaux2;
 return simbaux1;
}				
   

/* Partiendo de una ecuacion, devuelve una lista de todas las 
funciones con sus aridades, que aparecen en la parte izquierda
y derecha de la ecuacion */

PRIVATE struct funcs *ecufuncs(struct arbol *ecu, struct funcs **final)
{ 
 struct arbol *derecha, *izquierda;
 struct funcs *simbaux1=NULL, *simbaux2=NULL, *simbaux3=NULL;
 
 if (ecu!=NULL)
   {
     izquierda=ecu; derecha=ecu->herm; ecu->herm=NULL;
     simbaux1=termfuncs(izquierda, &simbaux2);
     if (simbaux1==NULL) 
       { simbaux1=termfuncs(derecha, &simbaux2); (*final)=simbaux2;}
     else 
       { simbaux2->sig=termfuncs(derecha, &simbaux3); 
	 if (simbaux2->sig) (*final)=simbaux3; else (*final)=simbaux2;}
     ecu->herm=derecha;
   }
 return simbaux1;
}


/* Partiendo de una regla, devuelve una lista de todas las 
funciones con sus aridades, que aparecen en la ecuacion
cabeza y en las posibles ecuaciones del cuerpo de la misma. */

PRIVATE struct funcs *reglafuncs(struct arbol *regla, struct funcs **final)
{ 
 struct arbol *ecuaux, *resto;
 struct funcs *simbaux1=NULL, *simbaux2=NULL, *simbaux3=NULL;
 
 if (regla!=NULL)
   {
     ecuaux=regla; resto=regla->herm->herm; regla->herm->herm=NULL;
     simbaux1=ecufuncs(ecuaux, &simbaux2);
     if (resto)
	{ if (simbaux1==NULL) 
	   { simbaux1=reglafuncs(resto, &simbaux2); (*final)=simbaux2; }
	  else
	   { simbaux2->sig=reglafuncs(resto, &simbaux3); 
	     if (simbaux2->sig) (*final)=simbaux3; else (*final)=simbaux2;}
	  regla->herm->herm=resto;
	}
     else (*final)=simbaux2;
   }
 return simbaux1;
}


/* Genera la lista de simbolos (signatura) de la teoria. Toma como
argumento una lista de reglas y la direccion de un puntero que al
final del proceso guardara la direccion de la cola de la lista de
simbolos. Devuelve un puntero a la cabeza de la lista de simbolos. */

PRIVATE struct funcs *gensigna(struct lista *ini, struct funcs **sbfin)
{
	struct lista *listaux1=NULL;
	struct funcs *simbaux1=NULL, /* cabeza lista simbolos */
		     *simbaux2=NULL, /* cola lista simbolos */
		     *simbaux3=NULL, /* puntero a simbolo a insertar */
		     *simbaux4=NULL; /* auxiliar */

	for(listaux1=ini; listaux1!=NULL; listaux1=listaux1->sig)
	{
	    simbaux3=reglafuncs(listaux1->ecu, &simbaux4);
	    for (;simbaux3!=NULL;)
		if (esta(simbaux1,simbaux3))
		  { simbaux4=simbaux3; simbaux3=simbaux3->sig; free(simbaux4);}
		else
			if (simbaux1==NULL)
			   { simbaux1=simbaux3; simbaux2=simbaux1;
			     simbaux3=simbaux3->sig; simbaux1->sig=NULL; }
			else
		    	   { simbaux2->sig=simbaux3; simbaux3=simbaux3->sig;
			    simbaux2=simbaux2->sig; simbaux2->sig=NULL;}
	};
	(*sbfin)= simbaux2;
	return simbaux1;
}


/* Comprueba errores en una lista de simbolos. Si aparecen dos
simbolos con igual nombre en la lista generada, sera indicio de
que estos simbolos se han utilizado en diferentes puntos de la
teoria con distinta aridad. Esto se reporta como error
sintactico. Es el ultimo filtro que debe superar la teoria en su
analisis sintactico. 
Recibe como argumento la cabeza a una lista de simbolos. 
Retorna 1 si hay error, 0 en caso contrario.
Esta funcion tambien se utilizara en el analisis de objetivos. */

PRIVATE int errsigna(struct funcs *sbini)
{
   struct funcs *simbaux1;
   for (simbaux1=sbini;simbaux1;simbaux1=simbaux1->sig)
	if (esta1(simbaux1->sig,simbaux1)) {
	  printf("El simbolo, [%s], se empleo con diferente aridad.\n",
							simbaux1->nombre);
	  return(1);
	};
   return(0);
}


/*  Detecta errores sintacticos en el objetivo introducido, no
achacables a la construccion de los terminos que lo componen.
Controla que una funcion se emplea con la aridad adecuada o, mas 
sencillamente, que el simbolo haya sido definido dentro de la teoria.
Genera una lista de simbolos asociados al objetivo y despues comprueba que
cada uno de estos simbolos esta en la lista de simbolos de la teoria
(funcini).
Se ha tomado la opcion de no ser tan riguroso en el tratamiento del
error. Simplemente se reporta. */

PRIVATE errgoal()
{
	struct funcs *sbgoalini, *sbgoalfin, *simbaux;

	/* generacion lista simbolos temporal asociada al objetivo
	actual. */
	sbgoalini=gensigna(goalini, &sbgoalfin);

	/* Primer filtro sintactico */
	errsigna(sbgoalini);

	/* Segundo filtro sintactico: Los simbolos deben estar definidos
	en la signatura introducida por la teoria. */
	for (simbaux=sbgoalini;simbaux;simbaux=simbaux->sig) 
	  if (!esta(funcini,simbaux))
	    printf("El simbolo [%s] no esta definido en la teoria.\n",
	    						simbaux->nombre);
	borrafun(sbgoalini);
}

/******************************************************************************
*
*		ANALISIS SINTACTICO DE TEORIAS Y OBJETIVOS.
*
*******************************************************************************/

/* Analisis sintactico de un conjunto de ecuaciones (teoria)
y generacion de la lista de simbolos */

PUBLIC  parser_teo()
{
   FILE *in;
  char filein[80], fileout[80];
  struct lista *listaux=NULL;

  /* Acciones para eliminar las estructuras relacionadas con la teoria anterior.  Se elimina la lista listaini, y la lista funcini. */
  i=0;
  cad[0]='\0'; semilla=0;
  if (listaini) { borralis(listaini); listaini=NULL; listafin=NULL;
		  borrafun(funcini); funcini=NULL; funcfin=NULL; };


  printf("\n\n\nENTRADA: Introduzca el nombre del fichero que contiene la teoria\n");
  printf("(Pulse RETURN si desea introducirla por teclado)\n:");
  gets(filein);
  if (filein[0])
   {
    if ((in=fopen(filein,"rt"))==NULL)
	{ printf("Error al abrir el fichero %s", filein); exit(1); };
    for (fgets((char *) cad,(LCAD-1),in),i=0;!feof(in);fgets((char *) cad,(LCAD-1),in),i=0)
					tratarteo(&listaini, &listafin);
    fclose(in);
   }
  else
   {
    printf("\nIntroduzca cada regla de la teoria finalizando con un return\n");
    printf("cada una de ellas. La ultima regla sera la vacia.\n");
    printf("El formato de una regla es: CABEZA <= CUERPO , donde CABEZA es \n");
    printf("una ecuacion, y CUERPO una secuencia (posiblemente vacia) de \n");
    printf("ecuaciones separadas por comas. \n\n");
    for (fgets((char *) cad,(LCAD-1),stdin),i=0;cad[0]!='\n';
	fgets((char *) cad,(LCAD-1),stdin),i=0) tratarteo(&listaini, &listafin);
   }

  /* Generacion de la lista de simbolos y ultimo filtro de
  errores sintacticos. */
  funcini=gensigna(listaini, &funcfin);
  if (errsigna(funcini)) exit(1); /* Opcion muy rigurosa, terminar. */

  /* Fichero estandar para la salida de resultados. */
  printf("\nSALIDA: Introduzca el nombre del fichero ");
  printf("donde se imprimiran los resultados\n");
  printf("(Pulse RETURN si desea que estos se muestren por pantalla)\n:");
  gets(fileout);
  if (fileout[0]==0) out=stdout;
  else if ((out=fopen(fileout,"wt"))==NULL)
	{ printf("Error al abrir el fichero %s", fileout); exit(1); }
       else /* Graba la teoria en el fichero de salida */
	verlis("\n\nTEORIA: \n", listaini);
}


PUBLIC parser_goal()
{
  /* Aqui habra que realizar acciones para borrar el goal
  anterior y todas las estructuras con el relacionadas. */
  i= 0;
  cad[0] = '\0';
  if (goalini) { borralis(goalini); goalini=NULL; goalfin=NULL;};

  printf("\nIntroduzca cada ecuacion del objetivo ajustandose\n");
  printf("a la siguiente sintaxis:\n\n");
  printf("\t\t<= ec1 , ec2, .. , ecN\n\n");
  printf("cada una de las ecuaciones separada por una '','' y\n");
  printf("finalizando el objetivo con un RETURN.\n\n");
  
  fgets((char *) cad, (LCAD-1), stdin);

  /* Analisis sintactico del objetivo y generacion de su lista de ecuaciones. */
  tratargoal(&goalini, &goalfin);

  /* Ultimo filtro para errores sintacticos */
  errgoal();
}


